/*----------------------------------------------------------------------
  File    : whopf.c
  Contents: Windows program for Hopfield networks as associative memory
  Author  : Christian Borgelt
  History : 2002.12.01 file created from file wsom.c
            2003.01.28 mouse event handling improved
            2014.10.24 changed from LGPL license to MIT license
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <time.h>
#include <windows.h>
#include "whfn.rh"
#include "hopfnet.h"
#include "random.h"

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME  "whfn"         /* main window title */

/* --- abbreviations --- */
#define HINST             HINSTANCE
#define MKIRSC            MAKEINTRESOURCE
#define SDIMSG            SendDlgItemMessage
#define WINSTYLE          WS_CAPTION|WS_SYSMENU|WS_BORDER|WS_MINIMIZEBOX
#define PATH_MAX          _MAX_PATH

/* --- edit box functions --- */
#define edt_set(h,i,s)    SetDlgItemText(h, i, s)
#define edt_get(h,i,s,n)  GetDlgItemText(h, i, s, n)
#define int_set(h,i,n)    SetDlgItemInt(h, i, n, 0)
#define int_get(h,i)      GetDlgItemInt(h, i, NULL, 0)

/* --- combo box functions --- */
#define cmb_add(h,i,s)    SDIMSG(h, i, CB_ADDSTRING,    0, (LPARAM)(s))
#define cmb_rem(h,i,k)    SDIMSG(h, i, CB_DELETESTRING, k, 0)
#define cmb_set(h,i,k)    SDIMSG(h, i, CB_SETCURSEL,    k, 0)
#define cmb_get(h,i)      SDIMSG(h, i, CB_GETCURSEL,    0, 0)

/* --- extensions --- */
#define INITXEXT     239        /* initial x-extension of window */
#define INITYEXT     288        /* initial y-extension of window */

/* --- error codes --- */
#define E_NONE         0        /* no error */
#define E_NOMEM      (-1)       /* not enough memory */
#define E_FOPEN      (-2)       /* cannot open file */
#define E_FREAD      (-3)       /* read error on file */
#define E_FWRITE     (-4)       /* write error on file */
#define E_INIT       (-5)       /* initialization failed */
#define E_PATCNT     (-6)       /* pattern capacity exceeded */
#define E_UNKNOWN    (-7)       /* unkown error */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
const char *errmsgs[] = {       /* error messages */
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file\n%s",
  /* E_FREAD    -3 */  "read error on file\n%s",
  /* E_FWRITE   -4 */  "write error on file\n%s",
  /* E_INIT     -5 */  "initialization failed",
  /* E_PATCNT   -6 */  "pattern capacity exceeded",
  /* E_UNKNOWN  -7 */  "unknown error"
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static HINST   hinst;           /* instance handle */
static HWND    hmain;           /* handle of main window */
static HDC     hdc_win;         /* device context of window */
static char    text[_MAX_PATH +1024];

/* --- pens and brushes --- */
static HPEN    pen_black;       /* black : grid of network */
static HBRUSH  br_white;        /* white : activation = -1 */
static HBRUSH  br_grey;         /* grey  : activation = +1 */

/* --- neural network information --- */
static struct {                 /* -- parameters -- */
  int    width, height;         /* width and height */
  int    grid;                  /* grid width */
  int    mode;                  /* update mode */
  int    curr;                  /* current pattern */
  double delay;                 /* delay (in seconds) */
} nnpar = { 10, 12, 24, 0, -1, 0.01 };

static HOPFNET *hfn    = NULL;  /* current reference vectors, */
static int     running = 0;     /* whether training is running */

/*----------------------------------------------------------------------
  Error Functions
----------------------------------------------------------------------*/

static void error (int code, ...)
{                               /* --- print error message */
  va_list     args;             /* list of variable arguments */
  const char  *msg;             /* error message */

  if ((code > 0) || (code < E_UNKNOWN))
    code = E_UNKNOWN;           /* check error code */
  va_start(args, code);         /* get variable arguments */
  msg = errmsgs[-code];         /* and error message text */
  if (!msg) msg = errmsgs[-E_UNKNOWN];
  vsprintf(text, msg, args);    /* format error message */
  MessageBox(hmain, text, PRGNAME, MB_OK|MB_ICONSTOP);
  va_end(args);                 /* show alert dialog box */
}  /* error() */

/*----------------------------------------------------------------------
  Drawing Functions
----------------------------------------------------------------------*/

static void redraw (HDC hdc)
{                               /* --- Actions > Redraw */
  RECT rc;                      /* rectangle for drawing */
  int  x, y;                    /* loop variable */
  int  g, a;                    /* grid width, activation */

  if (!hdc) hdc = hdc_win;      /* get the DC of the window */
  g = nnpar.grid;               /* get the grid width */
  for (y = hfn_height(hfn); --y >= 0; ) {
    for (x = hfn_width(hfn); --x >= 0; ) {
      a = hfn_actget(hfn, x,y); /* traverse the neurons */
      rc.left = x*g; rc.right  = rc.left +g-1;
      rc.top  = y*g; rc.bottom = rc.top  +g-1;
      FillRect(hdc, &rc, (a > 0) ? br_grey : br_white);
    }                           /* color each field according to */
  }                             /* the activation of the neuron */
  SelectObject(hdc, pen_black);
  y = hfn_height(hfn) *g -1;    /* draw horizontal grid lines */
  for (a = -1, x = hfn_width(hfn);  --x > 0; ) {
    a += g; MoveToEx(hdc, a, 0, NULL); LineTo(hdc, a, y); }
  x = hfn_width(hfn) *g -1;     /* draw vertical   grid lines */
  for (a = -1, y = hfn_height(hfn); --y > 0; ) {
    a += g; MoveToEx(hdc, 0, a, NULL); LineTo(hdc, x, a); }
}  /* redraw() */

/*--------------------------------------------------------------------*/

static void resize (HWND hwnd)
{                               /* --- adapt window size */
  RECT rc;                      /* client area rectangle */
  int  x, y, dx, dy;            /* coordinates and offsets */
  int  wx, wy;                  /* coordinates of window corner */

  GetClientRect(hwnd, &rc);     /* get client rectangle of window */
  rc.right  = rc.left +(dx = nnpar.width  *nnpar.grid -1);
  rc.bottom = rc.top  +(dy = nnpar.height *nnpar.grid -1);
  AdjustWindowRect(&rc, WINSTYLE, 1);
  x = rc.right  -rc.left;       /* get window client area and */
  y = rc.bottom -rc.top;        /* compute new window rectangle, */
  GetWindowRect(hwnd, &rc);     /* get the current window position */
  wx = rc.left; wy = rc.top;    /* and resize the window */
  MoveWindow(hwnd, wx, wy, x, y, 1);
  GetClientRect(hwnd, &rc);     /* get resulting rectangle */
  if ((rc.right  -rc.left != dx)     /* if resulting client area */
  ||  (rc.bottom -rc.top  != dy)) {  /* has a different size */
    x += dx -(rc.right  -rc.left);   /* (e.g. because of a */
    y += dy -(rc.bottom -rc.top);    /* two-line menu) */
    MoveWindow(hwnd, wx, wy, x, y, 1);
  }                             /* resize window a second time */
  InvalidateRect(hwnd, NULL,0); /* trigger window redraw */
}  /* resize() */

/*--------------------------------------------------------------------*/

static void mouse (int x, int y, int down)
{                               /* --- evaluate mouse movement */
  int  g;                       /* grid width */
  RECT rc;                      /* rectangle to fill */
  static int a = 0;             /* activation to set */

  if (running) return;          /* if update is running, abort */
  if (down < 0) { a = 0; return; }
  if (!down && (a == 0)) return;
  g = nnpar.grid;               /* compute the neuron coordinates */
  x /= g; if ((x < 0) || (x >= hfn_width(hfn)))  return;
  y /= g; if ((y < 0) || (y >= hfn_height(hfn))) return;
  if (down) {                   /* on mouse down, get activation */
    a = (hfn_actget(hfn, x, y) > 0) ? -1 : 1; }
  hfn_actset(hfn, x, y, a);     /* toggle the activation and redraw */
  rc.left = x*g; rc.right  = rc.left +g-1;
  rc.top  = y*g; rc.bottom = rc.top  +g-1;
  FillRect(hdc_win, &rc, (a > 0) ? br_grey : br_white);
}  /* mouse() */

/*--------------------------------------------------------------------*/

static void CALLBACK update (HWND hwnd, UINT msg, UINT id,
                             DWORD systime)
{                               /* --- do one training step */
  RECT rc;                      /* rectangle for drawing */
  int  x, y;                    /* loop variable */
  int  g, a;                    /* grid width, activation */

  if (!running) return;         /* check the update flag */
  if (nnpar.mode == 0) {        /* if in single neuron mode */
    hfn_curr(hfn, &x, &y);      /* get the neuron coordinates */
    a = hfn_update(hfn, -1, 0); /* and update the neuron */
    g = nnpar.grid;             /* redraw the corresp. field */
    rc.left = x*g; rc.right  = rc.left +g-1;
    rc.top  = y*g; rc.bottom = rc.top  +g-1;
    FillRect(hdc_win, &rc, (a > 0) ? br_grey : br_white); }
  else {                        /* if in epoch mode */
    hfn_update(hfn, -1, -1);    /* update all neurons */
    redraw(NULL);               /* redraw the window contents */
  }
  SetTimer(hwnd, 1, (UINT)(nnpar.delay *1000), (TIMERPROC)update);
}  /* update() */               /* reinstall timer */

/*----------------------------------------------------------------------
  Dialog Functions
----------------------------------------------------------------------*/

static char* fselect (char *title, int load)
{                               /* --- select a file */
  static char fname[_MAX_PATH+1];    /* file name */
  static char path [_MAX_PATH+1];    /* and path */
  static char filter[] =        /* file selection filter */
    "Hopfield networks (*.hfn)\0*.hfn\0"
    "all files (*.*)\0*.*\0";
  static OPENFILENAME ofn = {   /* file selector structure */
    sizeof(OPENFILENAME),       /* lStructSize */
    NULL,   NULL,               /* hwndOwner, hInstance */
    filter, NULL,               /* lpstrFilter, lpstrCustomFilter */
    0, 1,                       /* nMaxCustFilter, nFilterIndex */
    fname, _MAX_PATH,           /* lpstrFile, nMaxFile */
    NULL, _MAX_FNAME+_MAX_EXT,  /* lpstrFileTitle, nMaxFileTitle */
    path,                       /* lpstrInitialDir */
    "Select File...",           /* lpstrTitle */
    OFN_PATHMUSTEXIST, 0, 0,    /* Flags, nFileOffset, nFileExtension */
    "nn", 0L,                   /* lpstrDefExt, lCustData */
    NULL, NULL };               /* lpfnHook, lpTemplateName */

  ofn.hwndOwner  = hmain;       /* set window handle */
  ofn.lpstrTitle = title;       /* and title text */
  ofn.Flags = (load) ? OFN_FILEMUSTEXIST : OFN_PATHMUSTEXIST;
  if (GetOpenFileName(&ofn) == 0)
    return NULL;                /* open file selector box */
  return fname;                 /* return name of selected file */
}  /* fselect() */

/*--------------------------------------------------------------------*/

static int load (void)
{                               /* --- File > Load Network... */
  char    *fname;               /* name of file to load */
  FILE    *file;                /* input file */
  HOPFNET *tmp;                 /* temporary buffer for network */

  LoadString(hinst, CS_LOAD, text, 80);
  fname = fselect(text, 1);     /* get the dialog box title */
  if (!fname) return -1;        /* and select a file */
  file  = fopen(fname, "r");    /* open input file */
  if (!file) { error(E_FOPEN, fname); return -1; }
  tmp = hfn_load(file);         /* load the self-org. map */
  fclose(file);                 /* close the input file */
  if (!tmp) { error(E_FREAD, fname); return -1; }
  hfn_delete(hfn); hfn = tmp;   /* replace the existing network */
  nnpar.width  = hfn_width(hfn);
  nnpar.height = hfn_height(hfn);
  return 0;                     /* return 'ok' */
}  /* load() */

/*--------------------------------------------------------------------*/

static void save (void)
{                               /* --- File > Save Network... */
  char *fname;                  /* name of file to save to */
  FILE *file;                   /* output file */

  LoadString(hinst, CS_SAVE, text, 80);
  fname = fselect(text, 0);     /* get the dialog box title */
  if (!fname) return;           /* and select a file */
  file = fopen(fname, "w");     /* open output file */
  if (!file) { error(E_FOPEN, fname); return; }
  hfn_save(hfn, file);          /* save the Hopfield network */
  if (fclose(file) != 0) error(E_FWRITE);
}  /* save() */

/*--------------------------------------------------------------------*/

static void mn_enable (int enable)
{                               /* --- enable/disable menu items */
  HMENU        menu;            /* handle of window menu */
  MENUITEMINFO mii;             /* for menu customization */
  UINT         flag;            /* enable/disable flag */

  menu = GetMenu(hmain);        /* get menu handle */
  flag = (enable) ? MF_ENABLED : MF_GRAYED;
  EnableMenuItem(menu, MI_LOAD,   flag);  /* enable/disable context */
  EnableMenuItem(menu, MI_SAVE,   flag);  /* sensitive menu items */
  EnableMenuItem(menu, MI_GRID,   flag);
  EnableMenuItem(menu, MI_NEXT,   flag);
  EnableMenuItem(menu, MI_STORE,  flag);
  EnableMenuItem(menu, MI_DELETE, flag);
  mii.cbSize = sizeof(MENUITEMINFO);  /* set either the */
  mii.fMask  = MIIM_ID|MIIM_TYPE;     /* 'Start Training' or */
  mii.fType  = MFT_STRING;            /* 'Stop Training' item */
  mii.wID    = (enable) ? MI_START : MI_STOP;
  LoadString(hinst, (enable) ? CS_START : CS_STOP, text, 80);
  mii.dwTypeData = text; mii.cch = (UINT)strlen(text);
  SetMenuItemInfo(menu, (enable) ? MI_STOP : MI_START, FALSE, &mii);
}  /* mn_enable() */

/*----------------------------------------------------------------------
  Dialog Box Procedures
----------------------------------------------------------------------*/

void dbl_set (HWND hwnd, int item, double num)
{                               /* --- set floating point number */
  sprintf(text, "%g", num);     /* format floating point number */
  edt_set(hwnd, item, text);    /* set formatted number */
}  /* dbl_set() */

/*--------------------------------------------------------------------*/

double dbl_get (HWND hwnd, int item)
{                               /* --- get floating point number */
  edt_get(hwnd, item, text, 64);    /* get input text and */
  return strtod(text, NULL);    /* convert it to a number */
}  /* dbl_get() */

/*--------------------------------------------------------------------*/

BOOL CALLBACK dp_about (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{                               /* --- 'about' dialog box procedure */
  switch (msg) {                /* evaluate message */
    case WM_COMMAND:            /* --- if command */
      if (wp != IDOK) break;    /* process only 'Ok' button */
    case WM_CLOSE:              /* --- if dialog closed */
      EndDialog(hwnd, 0);       /* terminate dialog */
      return 1;                 /* return 'ok' */
  }
  return 0;                     /* everything else not handled */
}  /* dp_about() */

/*--------------------------------------------------------------------*/

BOOL CALLBACK dp_grid (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{                               /* --- grid/shape dialog procedure */
  switch (msg) {                /* evaluate message */
    case WM_INITDIALOG:         /* dialog initialization */
      int_set(hwnd, DI_WIDTH,  nnpar.width);
      int_set(hwnd, DI_HEIGHT, nnpar.height);
      int_set(hwnd, DI_GRID,   nnpar.grid);
      break;
    case WM_COMMAND:            /* command */
      switch (wp) {             /* evaluate command code */
        case IDOK:              /* if 'Ok' button pressed */
          nnpar.width  = int_get(hwnd, DI_WIDTH);
          if      (nnpar.width  < 2)   nnpar.width  = 2;
          else if (nnpar.width  > 100) nnpar.width  = 100;
                                /* width  of network (in units) */
          nnpar.height = int_get(hwnd, DI_HEIGHT);
          if      (nnpar.height < 1)   nnpar.height = 1;
          else if (nnpar.height > 100) nnpar.height = 100;
                                /* height of network (in units) */
          nnpar.grid   = int_get(hwnd, DI_GRID);
          if      (nnpar.grid   < 4)   nnpar.grid   = 4;
          else if (nnpar.grid   > 64)  nnpar.grid   = 64;
                                /* grid width (in pixels) */
          EndDialog(hwnd, 0);   /* get inputs from edit controls */
          return 1;             /* return 'ok' */
        case IDCANCEL:          /* if 'Cancel' button pressed, */
          EndDialog(hwnd, -1); return 1;   /* terminate dialog */
      } break;
    case WM_CLOSE:              /* dialog box has been closed */
      EndDialog(hwnd, -1); return 1;      /* terminate dialog */
  }
  return 0;                     /* everything else not handled */
}  /* dp_grid() */

/*--------------------------------------------------------------------*/

BOOL CALLBACK dp_params (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{                               /* --- parameters dialog procedure */
  switch (msg) {                /* evaluate message */
    case WM_INITDIALOG:         /* dialog initialization */
      LoadString(hinst, CS_SINGLE, text, 80);
      cmb_add(hwnd, DI_MODE,   text);
      LoadString(hinst, CS_ALL,    text, 80);
      cmb_add(hwnd, DI_MODE,   text);
      cmb_set(hwnd, DI_MODE,   nnpar.mode);
      dbl_set(hwnd, DI_DELAY,  nnpar.delay);
      break;                    /* set learning parameters */
    case WM_COMMAND:            /* command */
      switch (wp) {             /* evaluate command code */
        case IDOK:              /* if 'Ok' button pressed */
          nnpar.mode  = (int)cmb_get(hwnd, DI_MODE);
          nnpar.delay = dbl_get(hwnd, DI_DELAY);  /* update mode */
          if      (nnpar.delay < 0.001) nnpar.delay  = 0.001;
          else if (nnpar.delay > 10)    nnpar.delay  = 10;
                                /* delay between redraws */
          EndDialog(hwnd, 0);   /* terminate dialog and */
          return 1;             /* return 'ok' */
        case IDCANCEL:          /* if 'Cancel' button pressed, */
          EndDialog(hwnd, -1); return 1;   /* terminate dialog */
      } break;
    case WM_CLOSE:              /* dialog box has been closed */
      EndDialog(hwnd, -1); return 1;      /* terminate dialog */
  }
  return 0;                     /* everything else not handled */
}  /* dp_params() */

/*----------------------------------------------------------------------
  Window Procedure
----------------------------------------------------------------------*/

LRESULT CALLBACK wp_main (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{                               /* --- main window procedure */
  static PAINTSTRUCT ps;        /* for BeginPaint call */
  HDC     hdc;                  /* handle of device context */
  int     i;                    /* pattern index */
  HOPFNET *tmp;                 /* buffer for new Hopfield network */

  switch (msg) {                /* evaluate message */
    case WM_PAINT:              /* --- if to paint window */
      hdc = BeginPaint(hwnd, &ps);
      if (!hdc) return 0;       /* get device context for drawing */
      redraw(hdc);              /* redraw the window contents */
      EndPaint(hwnd, &ps);      /* and release the device context */
      break;

    case WM_COMMAND:            /* --- if menu item selected */
      switch (wp) {             /* evaluate item */
        case MI_LOAD:           /* Menu Item 'Load Map...' */
          if (load() == 0) resize(hwnd);
          return 0;             /* load a map and redraw the window */

        case MI_SAVE:           /* Menu Item 'Save Map...' */
          save(); return 0;     /* save a self-organizing map */

        case MI_QUIT:           /* Menu Item 'Quit' */
          DestroyWindow(hwnd);  /* stop network training */
          return 0;             /* and destroy the window */

        case MI_CLEAR:          /* Menu Item 'Clear' */
          hfn_clear(hfn);       /* clear all activations */
          redraw(NULL);         /* and redraw the window */
          return 0;

        case MI_NEXT:           /* Menu Item 'Next Pattern' */
          if (running) return 0;/* if update is running, abort */
          i = hfn_patcnt(hfn);  /* get and check */
          if (i <= 0) return 0; /* the number of patterns */
          if (++nnpar.curr >= i) nnpar.curr = 0;
          hfn_patget(hfn, nnpar.curr);
          redraw(NULL);         /* get the current pattern */
          return 0;             /* and redraw the window */

        case MI_STORE:          /* Menu Item 'Store Pattern' */
          if (running) return 0;/* if update is running, abort */
          if (hfn_patadd(hfn) != 0)
            error(E_PATCNT);    /* store the current pattern */
          return 0;             /* and check for an error */

        case MI_DELETE:         /* Menu Item 'Delete Pattern' */
          if (running) return 0;/* if update is running, abort */
          i = hfn_patfind(hfn);
          if (i >= 0) hfn_patrem(hfn, i);
          return 0;             /* find and delete the pattern */

        case MI_INIT:           /* Menu Item 'Initialize' */
          hfn_rand(hfn, drand); /* set random activations */
          redraw(NULL);         /* and redraw the window */
          return 0;

        case MI_START:          /* Menu Item 'Start Training' */
          mn_enable(0);         /* disable certain menu items */
          running = 1; update(hmain, WM_TIMER, 1, 0);
          return 0;             /* start the training */

        case MI_STOP:           /* Menu Item 'Stop Training' */
          KillTimer(hmain, 1);  /* kill the timer */
          mn_enable(1);         /* enable all menu items */
          return running = 0;   /* clear the running flag */

        case MI_REDRAW:         /* Menu Item 'Redraw' */
          InvalidateRect(hwnd, NULL, 0);
          return 0;             /* invalidate window contents */

        case MI_GRID:           /* Menu Item 'Grid/Shape...' */
          if (DialogBox(hinst, MKIRSC(DB_GRID), hwnd,
                       (DLGPROC)dp_grid) == 0) {
            tmp = hfn_create(nnpar.width, nnpar.height);
            if (!tmp) { error(E_NOMEM); return 0; }
            hfn_delete(hfn); hfn = tmp;
            resize(hwnd);       /* replace the Hopfield network */
          }                     /* and redraw the window contents */
          return 0;             /* get pattern region and grid size */

        case MI_PARAMS:         /* Menu Item 'Parameters...' */
          if (DialogBox(hinst, MKIRSC(DB_PARAMS), hwnd,
             (DLGPROC)dp_params) == 0)
            InvalidateRect(hwnd, NULL, 0);
          return 0;             /* get learning parameters */

        case MI_ABOUT:          /* Menu Item 'About' */
          DialogBox(hinst, MKIRSC(DB_ABOUT), hwnd, (DLGPROC)dp_about);
          return 0;             /* show dialog box 'About Wsom' */
      } break;

    case WM_LBUTTONDOWN:        /* left mouse button pressed */
      mouse(LOWORD(lp), HIWORD(lp), 1);
      return 0;                 /* process the mouse click */

    case WM_LBUTTONUP:          /* left mouse button released */
      mouse(LOWORD(lp), HIWORD(lp), -1);
      return 0;                 /* process the mouse click */

    case WM_MOUSEMOVE:          /* mouse moved */
      mouse(LOWORD(lp), HIWORD(lp), 0);
      return 0;                 /* process the mouse click */

    case WM_CLOSE:              /* window closed */
    case WM_DESTROY:            /* or destroyed */
      PostQuitMessage(0);       /* terminate program */
      return 0;
  }
  return DefWindowProc(hwnd, msg, wp, lp);
}  /* wp_main() */

/*----------------------------------------------------------------------
  Main Program
----------------------------------------------------------------------*/

int WINAPI WinMain (HINSTANCE hcurr, HINSTANCE hprev,
                    LPSTR cmdln, int cmd)
{                               /* --- main function */
  WNDCLASS wc;                  /* window class */
  RECT     rc;                  /* window rectangle */
  MSG      msg;                 /* message */

  /* --- create window class --- */
  hinst = hcurr;                /* note instance handle */
  if (!hprev) {                 /* if first instance of this program */
    wc.style         = CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc   = wp_main;
    wc.cbClsExtra    = wc.cbWndExtra = 0;
    wc.hInstance     = hinst;
    wc.hIcon         = LoadIcon(NULL, IDI_APPLICATION);
    wc.hCursor       = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = NULL;
    wc.lpszMenuName  = MKIRSC(MN_MAIN);
    wc.lpszClassName = "WHfNClass";
    if (!RegisterClass(&wc))    /* register class */
      return 0;                 /* of main window */
  }
  srand((unsigned long)time(NULL));
  hfn = hfn_create(nnpar.width, nnpar.height);
  if (!hfn) error(E_NOMEM);

  /* --- get pens and brushes --- */
  pen_black = GetStockObject(BLACK_PEN);
  br_white  = GetStockObject(WHITE_BRUSH);
  br_grey   = CreateSolidBrush(RGB(0x40, 0x40, 0x40));

  /* --- create window --- */
  rc.left = 0; rc.right  = nnpar.width  *nnpar.grid -2;
  rc.top  = 0; rc.bottom = nnpar.height *nnpar.grid -2;
  AdjustWindowRect(&rc, WINSTYLE, 1);
  hmain = CreateWindow("WHfNClass", PRGNAME, WINSTYLE,
                       30, 30, rc.right-rc.left, rc.bottom-rc.top,
                       NULL, NULL, hinst, NULL);
  if (!hmain) return 0;         /* create main window */
  hdc_win = GetDC(hmain);       /* get a device context */
  if (!hdc_win) error(E_INIT);  /* for independent drawing */
  ShowWindow(hmain, cmd);       /* open and */
  UpdateWindow(hmain);          /* update main window */

  /* --- dispatch messages --- */
  while (GetMessage(&msg, NULL, 0, 0) > 0) {
    TranslateMessage(&msg);     /* get messages, translate */
    DispatchMessage (&msg);     /* and dispatch them */
  }

  /* --- clean up --- */
  ReleaseDC(hmain, hdc_win);    /* release device context of window */
  DeleteObject(br_grey);        /* delete created brush */
  return (int)msg.wParam;
}  /* WinMain() */
